/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __NNFW_KERNEL_ACL_DEPTHWISECONV2D_COMMON_H__
#define __NNFW_KERNEL_ACL_DEPTHWISECONV2D_COMMON_H__

#include <OperationsUtils.h>
#include <arm_compute/core/TensorShape.h>
#include <arm_compute/core/TensorInfo.h>
#include <arm_compute/runtime/IFunction.h>

#include "shape.h"
#include "IO_accessor.h"

namespace nnfw {
namespace kernel {
namespace acl {

namespace common {

typedef std::function<void (void)> sync_scheduler_f;

template<class TensorT, class LayerT, class ActT>
bool depthwiseConvFloat32(const float* inputData, const nnfw::rt::Shape& inputShape,
                          const float* filterData, const nnfw::rt::Shape& filterShape,
                          const float* biasData, const nnfw::rt::Shape& biasShape,
                          int32_t padding_left, int32_t padding_right,
                          int32_t padding_top, int32_t padding_bottom,
                          int32_t stride_width, int32_t stride_height,
                          int32_t depth_multiplier, int32_t activation,
                          float* outputData, const nnfw::rt::Shape& outputShape,
                          sync_scheduler_f sync_scheduler) {
  auto inputShapeACL = util::fromNNShape(inputShape);
  auto weightsShapeACL = util::fromNNShape(filterShape);
  auto biasShapeACL = util::fromNNShape(biasShape);
  auto outputShapeACL = util::fromNNShape(outputShape);

  TensorT input(arm_compute::TensorInfo(inputShapeACL, arm_compute::Format::F32));
  TensorT weights(arm_compute::TensorInfo(weightsShapeACL, arm_compute::Format::F32));
  TensorT bias(arm_compute::TensorInfo(biasShapeACL, arm_compute::Format::F32));
  TensorT output(arm_compute::TensorInfo(outputShapeACL, arm_compute::Format::F32));

  arm_compute::PadStrideInfo psinfo = arm_compute::PadStrideInfo(stride_width, stride_height,
                                              padding_left, padding_right,
                                              padding_top, padding_bottom,
                                              arm_compute::DimensionRoundingType::FLOOR);

  auto l = std::make_shared<LayerT>();
  l->configure(input.ptr(), weights.ptr(), bias.ptr(), output.ptr(), psinfo);

  std::vector<std::shared_ptr<arm_compute::IFunction>> fns;

  fns.emplace_back(l);

  util::insertFusedActivationLayer<TensorT, ActT>(output, activation, fns);

  input.allocate();
  output.allocate();
  bias.allocate();
  weights.allocate();

  // TODO: Do we need 2D tensor accessor for the input feature?
  TensorAccess<InputAccessor>(input.ref(), inputData, inputShape);
  TensorAccess<BiasAccessor>(bias.ref(), biasData, biasShape);
  TensorAccess<WeightAccessor>(weights.ref(), filterData, filterShape);

  for (const auto &fn : fns)
  {
    fn->run();
  }

  sync_scheduler();

  TensorAccess<OutputAccessor>(output.ref(), outputData, outputShape);

  return true;
}

} // namespace common

} // namespace acl
} // namespace kernel
} // namespace nnfw

#endif // __NNFW_KERNEL_ACL_DEPTHWISECONV2D_COMMON_H__
