/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <util/fp32.h>

#include "util.h"

namespace nnfw {
namespace kernel {
namespace acl {
namespace util {

void initData(float* data, int num, float value)
{
  for (int i = 0; i < num; i++) {
    *(data + i) = value;
  }
}

void initData_Increasing(float* data, int num, float value)
{
  for (int i = 0; i < num; i++) {
    *(data + i) = value;
    value++;
  }
}

// compareData
// return true if result == expected with the shape info,
// otherwise false
bool compareData(const float* result, const float* expected, const nnfw::rt::Shape& shape)
{
  if (shape.dimensions.size() == 4)
  {
  // TODO fix indentation
  uint32_t height = nnfw::rt::getSizeOfDimension(shape, 1);
  uint32_t width  = nnfw::rt::getSizeOfDimension(shape, 2);
  uint32_t numitems = height * width;
  for (int item = 0; item < numitems; item++) {
    if (!::nnfw::util::fp32::epsilon_equal(*(result + item), *(expected + item), 1)) {
      LOG(ERROR) << "compareData failed: result " << *(result + item)
                 << ", expected " << *(expected + item)
                 << ", diff " << ::nnfw::util::fp32::relative_diff(*(result + item), *(expected + item))
                 << std::endl;
      return false;
    }
  }
  }
  else if (shape.dimensions.size() == 2)
  {
    uint32_t height = nnfw::rt::getSizeOfDimension(shape, 0);
    uint32_t width  = nnfw::rt::getSizeOfDimension(shape, 1);
    uint32_t numitems = height * width;
    for (int item = 0; item < numitems; item++) {
      if (!::nnfw::util::fp32::epsilon_equal(*(result + item), *(expected + item), 1)) {
        LOG(ERROR) << "compareData failed: result " << *(result + item)
                   << ", expected " << *(expected + item)
                   << ", diff " << ::nnfw::util::fp32::relative_diff(*(result + item), *(expected + item))
                   << std::endl;
        return false;
      }
    }
  }
  else
  {
    // TODO: add a handler for rank 1 and 3
    LOG(ERROR) << "Unhandled shape: " << shape.dimensions.size() << std::endl;
  }
  return true;
}

void NCHW2NHWC(const float* nchw, float* nhwc, const nnfw::rt::Shape& shape)
{
  uint32_t N = nnfw::rt::getSizeOfDimension(shape, 0);
  uint32_t H = nnfw::rt::getSizeOfDimension(shape, 1);
  uint32_t W = nnfw::rt::getSizeOfDimension(shape, 2);
  uint32_t C = nnfw::rt::getSizeOfDimension(shape, 3);

  for (uint32_t n = 0; n < N; n++) {
    for (uint32_t c = 0; c < C; c++) {
      for (uint32_t h = 0; h < H; h++) {
        for (uint32_t w = 0; w < W; w++) {
          uint32_t soffset = w + (h * W) + (c * W * H) + (n * W * H * C);
          uint32_t doffset = c + (w * C) + (h * C * W) + (n * C * W * H);
          *(nhwc + doffset) = *(nchw + soffset);
        }
      }
    }
  }
}

} // namespace util
} // namespace acl
} // namespace kernel
} // namespace nnfw
